#include "ContigFattener.hh"

ContigFattener::ContigFattener(bool verbose)
{
  m_verbose = verbose;
  m_contig = NULL;
  aligner = new SWAlign(verbose);
}

ContigFattener::~ContigFattener()
{
  delete aligner;
}

int ContigFattener::extendLeft(ContigSequence * read, int distance)
{
  int retval = 0;
  string leftTrim = read->getLeftTrim();
  if (distance == -1) { distance = leftTrim.size(); }

  if (distance > (int)leftTrim.size())
  {
    cerr << "distance(" << distance 
         << ") > leftTrim(" << leftTrim.size() << endl;

    throw Exception("distance > leftTrim.size()", "extendLeft");
  }

  int origoffset = read->getOffset();
  if (origoffset != 0)
  {
    // We need to fatten this read to the left end of the contig
    int distanceAligned = aligner->fattenLeft(m_contig, read, distance);
    m_gindex = m_contig->getWidth();

    retval += distanceAligned;

    distance -= distanceAligned;
    leftTrim.resize(leftTrim.size()-distanceAligned);

    m_contig->stripSequenceEdgeGaps();
  }

  if (distance)
  {
    int newoffset = read->getOffset();

    if (newoffset != 0)
    {
      if (m_verbose)
      {
        cerr << "fattening failed to reach the end of the consensus" << endl;
      }
    }

    if (newoffset - distance < 0)
    {
      // need to shift the entire contig right
      int contigShiftDistance = distance - newoffset;

      m_gindex = m_contig->shiftContigRight(contigShiftDistance);
    }

    newoffset = read->getOffset(); // shifted as necessary

    // Now add the read into the tiling
    m_contig->addReadToTiling(read, 0, distance);

    // Trim off the left side (3' side) that isn't wanted
    long int eraseDistance = leftTrim.size() - distance;
    if (eraseDistance)
    {
      leftTrim.erase(0, eraseDistance);
    }

    // Finally, extend the read
    retval += read->extendLeft(leftTrim);

    m_contig->dumpRange(read->getOffset()-10, newoffset+10, read);
  }

  return retval;
}

int ContigFattener::extendRight(ContigSequence * read, int distance)
{
  int retval = 0;
  string rightTrim = read->getRightTrim();
  if (distance == -1) { distance = rightTrim.size(); }

  if (distance > (int)rightTrim.size())
  {
    cerr << "distance(" << distance 
         << ") > rightTrim(" << rightTrim.size() << endl;

    throw Exception("distance > rightTrim.size()", "extendRight");
  }


  int origend = read->getRightEnd();
  if (origend != m_gindex-1)
  {
    // We need to fatten this read to the right end of the contig
    int distanceAligned = aligner->fattenRight(m_contig, read, distance);
    retval += distanceAligned;
    m_gindex = m_contig->getWidth();

    distance -= distanceAligned;
    rightTrim.erase(0, distanceAligned);

    m_contig->stripSequenceEdgeGaps();
  }

  if (distance)
  {
    int newend = read->getRightEnd();

    if (newend != m_gindex-1)
    {
      cerr << "fattening failed to reach the end of the consensus" << endl;
    }

    if (newend + distance >= m_gindex)
    {
      // open space in the consensus and tiling
      // need to grow contig
      int contigGrowDistance =  newend + distance + 1 - m_gindex;

      m_gindex = m_contig->widenContig(contigGrowDistance);
    }

    // Now add the read into the tiling
    m_contig->addReadToTiling(read, read->getRightEnd()+1, distance);

    // Finally, extend the read
    rightTrim.resize(distance);
    retval += read->extendRight(rightTrim);

    m_contig->dumpRange(origend-10, read->getRightEnd()+10, read);
  }

  return retval;
}

void ContigFattener::recordExtension(const ContigSequence * read, 
                                     int distanceExtended,
                                     int totalTrim,
                                     float origAvg,
                                     float origTrimAvg,
                                     int origUContigWidth)
{
  bool rc = read->isReverseCompliment();
  int offset = read->getOffset();
  int rightEnd = read->getRightEnd();
  int count = 10;

  float avg = read->getAvgQV((rc) ? offset : rightEnd, count);
  float trimavg = read->getAvgQV3Trim(count);

  int distanceToEnd = (rc) ? offset : m_gindex - rightEnd-1;
  int deltaUContig = m_contig->getUngappedIndex(m_gindex-1) - origUContigWidth;

  fprintf(stderr, "%s=%d %d %d %d %0.1f %0.1f %0.1f %0.1f\n",
          read->getSeqname().c_str(),
          totalTrim,
          distanceExtended,
          deltaUContig,
          distanceToEnd,
          origAvg,
          origTrimAvg,
          avg,
          trimavg);
}

void ContigFattener::initializePublicAPI(Contig * contig)
{
  m_contig = contig;
  m_gindex = m_contig->getWidth();
}


// Public API
void ContigFattener::addRead(Contig * contig, const string & seqname)
{
  initializePublicAPI(contig);

  throw Exception("Not implemented!", "ContigFattener::addRead");
}


// Public API
void ContigFattener::extendRead(Contig * contig,
                                const string & seqname, 
                                int distance)
{
  initializePublicAPI(contig);


  ContigSequence * read = m_contig->getContigSequence(seqname);
  if (read)
  {
    int distanceExtended = 0;
    int trimLength = read->get3TrimLength();
    int origUContigWidth = m_contig->getUngappedIndex(m_gindex-1);

    if (m_verbose)
    {
      cerr << "Extending r" << read->getReadCount() 
           << " by " << distance 
           << " of " << trimLength << endl;
    }

    if (trimLength)
    {
      if (read->isReverseCompliment())
      {
        distanceExtended = extendLeft(read, distance);
      }
      else
      {
        distanceExtended = extendRight(read, distance);
      }
    }

    m_contig->recallCons();
    
    recordExtension(read, distanceExtended, trimLength, 
                    0.0, 0.0, origUContigWidth);
  }
  else
  {
    throw Exception("Specified read (" + seqname + ") not found", 
                    "ContigFattener::extendRead");
  }
}

// Public API
void ContigFattener::growContigLeft(Contig * contig, bool requireConfirmation)
{
  initializePublicAPI(contig);

  // First try to fatten the left side
  fattenContig(contig, 0, 500, 1, 0);
  int confirmedDistance = 0;

  // Find the read that now extends the furthest
  int bestExtension = 0;
  ContigSequence * best = NULL;
  ContigSequenceSlice::const_iterator t;
  ContigSequenceSlice & tilingSlice = m_contig->getTilingSlice(0);

  for (t =  tilingSlice.begin();
       t != tilingSlice.end();
       t++)
  {
    if ((*t)->isReverseCompliment())
    {
      int currentExtension = (*t)->get3TrimLength();
      if (currentExtension > bestExtension)
      {
        bestExtension = currentExtension;
        best = *t;
      }
    }
  }

  // Stretch the contig with the best candidate
  if (best)
  {
    if (m_verbose)
    {
      cerr << "best read to left extend is r" << best->getReadId()
           << " by " << bestExtension << "bp" << endl;
    }

    // best->offset == 0, so this extended by exactly bestExtension bp
    extendRead(contig, best->getSeqname(), -1);

    // Now fatten the contig to the new beginning
    int origSize = m_gindex;
    fattenContig(contig, 0, bestExtension+500, 1, 0);
    bestExtension += (m_gindex - origSize); // add extra gaps

    for (int gindex = bestExtension-1; gindex >= 0; gindex--)
    {
      ContigSequenceSlice & tilingSlice = m_contig->getTilingSlice(gindex);
      if (tilingSlice.size() > 1)
      {
        confirmedDistance++;
      }
      else
      {
        if (requireConfirmation && 
            (tilingSlice.begin() != tilingSlice.end()) &&
            (*tilingSlice.begin())->getReadCount() == best->getReadCount())
        {
          // We need to trim from 0 to gindex since these are unconfirmed
          best->trimReadLeft(gindex+1);
          m_gindex = m_contig->trimContigLeft(gindex+1);
          m_contig->dumpRange(best->getOffset(), bestExtension+10, best);
        }

        break;
      }
    }
  }

  cerr << "@ConfirmLeftContigGrowDistance=" << confirmedDistance << endl;
}

// Public API
void ContigFattener::growContigRight(Contig * contig, bool requireConfirmation)
{
  initializePublicAPI(contig);

  int confirmedDistance = 0;

  // First try to grow the right side
  fattenContig(contig, m_gindex-500, m_gindex, 0, 1);

  // Find the read that now extends the furthest
  int bestExtension = 0;
  ContigSequence * best = NULL;
  ContigSequenceSlice::const_iterator t;
  ContigSequenceSlice & tilingSlice = m_contig->getTilingSlice(m_gindex-1);

  for (t =  tilingSlice.begin();
       t != tilingSlice.end();
       t++)
  {
    if (!(*t)->isReverseCompliment())
    {
      int currentExtension = (*t)->get3TrimLength();
      if (currentExtension > bestExtension)
      {
        bestExtension = currentExtension;
        best = *t;
      }
    }
  }

  // Stretch the contig with the best candidate
  if (best)
  {
    int oldSize = m_gindex;
    if (m_verbose)
    {
      cerr << "best read to right extend is r" << best->getReadId()
           << " by " << bestExtension << "bp" << endl;
    }

    extendRead(contig, best->getSeqname(), -1);
    
    // Now fatten the contig to match out to the new end
    fattenContig(contig, oldSize-1, m_gindex, 0, 1);

    for (int i = oldSize; i < m_gindex; i++)
    {
      ContigSequenceSlice & tilingSlice = m_contig->getTilingSlice(i);
      if (tilingSlice.size() > 1)
      {
        confirmedDistance++;
      }
      else
      {
        if (requireConfirmation && 
            (tilingSlice.begin() != tilingSlice.end()) &&
            (*tilingSlice.begin())->getReadCount() == best->getReadCount())
        {
          // We need to trim from i to m_gindex since these are unconfirmed
          m_gindex = m_contig->trimContigRight(i);
          best->trimReadRight(i);
        }

        break;
      }
    }
  }

  cerr << "@ConfirmRightContigGrowDistance=" << confirmedDistance << endl;
}

// Public API
void ContigFattener::fattenContig(Contig * contig,
                                  int fattenRangeStart, int fattenRangeEnd,
                                  int doLeftExtension, int doRightExtension)
{
  initializePublicAPI(contig);

  int contigSize = m_gindex-1;
  int origUContigWidth = m_contig->getUngappedIndex(contigSize);

  if (fattenRangeEnd > contigSize)
  {
    fattenRangeEnd = contigSize;
  }

  if (m_verbose)
  {
    if (doLeftExtension)
    {
      cerr << "Left ";
      if (doRightExtension) { cerr << "and "; }
    }

    if (doRightExtension) { cerr << "Right "; }

    cerr << "Fattening in [" << fattenRangeStart
         << "," << fattenRangeEnd
         << "]" << endl;
  }

  ReadVector & m_readVector = m_contig->getReadVector();

  ReadVector::iterator i;
  for (i =  m_readVector.begin();
       i != m_readVector.end();
       i++)
  {
    ContigSequence * read = *i;
    int fattenRead = 0;
    int distanceExtended = 0;
    int trimLength = read->get3TrimLength();

    int count = 10;
    float origAvg = 0.0;
    float origTrimAvg = read->getAvgQV3Trim(count);

    if (read->isReverseCompliment())
    {
      origAvg = read->getAvgQV(read->getOffset(), count);

      fattenRead = doLeftExtension;
      if (fattenRead && fattenRangeStart != -1)
      {
        int offset = read->getOffset();
        fattenRead = fattenRangeStart <= offset && offset <= fattenRangeEnd;
      }

      if (fattenRead && trimLength)
      {
        distanceExtended = aligner->fattenLeft(m_contig, read, -1);
        m_gindex = m_contig->getWidth();
      }
    }
    else
    {
      origAvg = read->getAvgQV(read->getRightEnd(), count);

      fattenRead = doRightExtension;
      if (fattenRead && fattenRangeStart != -1)
      {
        int right = read->getRightEnd();
        fattenRead = fattenRangeStart <= right && right <= fattenRangeEnd;
      }

      if (fattenRead && trimLength)
      {
        distanceExtended = aligner->fattenRight(m_contig, read, -1);
        m_gindex = m_contig->getWidth();
      }
    }

    if (fattenRead)
    {
      // TODO: Fix the range when gaps are inserted outside of the fatten range. 
      // Fortunately this will only falsely increase the range to fatten
      
      // Shift the end of the range to reflect the gaps inserted in the fatten
      // range
      if (fattenRangeEnd != -1)
      {
        int distanceToEnd = contigSize - fattenRangeEnd;
        contigSize = m_gindex-1;
        fattenRangeEnd = contigSize - distanceToEnd;
      }

      recordExtension(read, distanceExtended, trimLength, 
                      origAvg, origTrimAvg, origUContigWidth);
      origUContigWidth = m_contig->getUngappedIndex(m_gindex-1);
    }
  }
}
