#include "SliceRotator.hh"
#include <cstdio>

SliceRotator::SliceRotator() : RequestManager("SliceRotator")
{
  m_log.debugL("Contructed SliceRotator");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);

  m_readCount = 0;
  m_newIndex = 0;
  m_newGindex = 0;
  m_firstSlice = true;
}

int SliceRotator::m_linear(0);
int SliceRotator::m_rotateSize(0);

SliceRotator::~SliceRotator()
{
  m_log.debugL("Destructed SliceRotator");
}


void SliceRotator::addRequest(ConsensusRequest * request)
{
  m_requestData = request;
  printXMLHeader(request);

  m_out << "  <ReadCollection>" << endl;
}

void SliceRotator::addReadData(ReadData * readData)
{
  readData->printXML(m_out);
}

void SliceRotator::addSliceRange(long & index, long & gindex)
{
  m_origSRIndex = index;
  m_origSRGindex = gindex;
  index = gindex = 0L;
}

void SliceRotator::printSlice(SliceData * slice)
{
  char consensus = slice->getConsensus();
  long coordinate = m_newIndex;
  if (coordinate && consensus == '-') coordinate--;
  slice->resetIndex(coordinate, m_newGindex);
  slice->printXML(m_out);

  if (consensus != '-') m_newIndex++;
  m_newGindex++;
}


void SliceRotator::addSlice(SliceData * slice)
{
  if (m_firstSlice)
  {
    m_newIndex = slice->getIndex();
    m_newGindex = slice->getGindex();

    m_firstSlice = false;
  }

  if (m_rotateSize > 0)
  {
    // Add the slice to the rotation queue
    // Don't do anything special about the readids here
    m_rotationQueue.push_back(slice);
    m_rotateSize--;
  }
  else
  {
    if (m_rotateSize == 0)
    {
      if (m_linear && !m_rotationQueue.empty())
      {
        // figure out which reads are broken by the new origin
        SliceData * lastSlice = m_rotationQueue.back();
        typedef set<string> ReadSet;
        ReadSet lastReadSet;

        // find the reads in the last rotated slice
        unsigned short depth = lastSlice->getDepth();
        for (unsigned short i = 0; i < depth; i++)
        {
          string seqid = lastSlice->getRead(i)->getId();
          lastReadSet.insert(seqid);
        }

        // check each read in the first non-rotated slice
        depth = slice->getDepth();
        for (unsigned short i = 0; i < depth; i++)
        {
          const ReadData * rd = slice->getRead(i);
          string seqid = rd->getId();

          if (lastReadSet.find(seqid) != lastReadSet.end())
          {
            // Read was split at the origin
            string seqname = rd->getSeqname();
            seqid += "'";
            seqname += "'";

            if (m_readManager->find(seqid) == NULL)
            {
              // load the read as read'
              ReadData * newRead;
              newRead = m_readManager->addReadData(seqid.c_str(),
                                                   seqname.c_str(), 
                                                   rd->isReverseCompliment() ? 
                                                     "1" : "0", 
                                                   rd->getChem().c_str());

              newRead->printXML(m_out);
            }
          }
        }
      }

      // Finish the readcollection
      
      m_out << "  </ReadCollection>" << endl;

      m_out << "  <SliceRange Index=\"" << m_origSRIndex
            << "\" Gindex=\""           << m_origSRGindex
            << "\">" << endl;

      m_rotateSize--;
    }

    // m_rotateSize == -1

    // Just print the slices
    printSlice(slice);
    delete slice;
  }
}

int SliceRotator::printResults()
{
  // Print the slices in the rotation queue
  
  SliceList::const_iterator s;

  for (s =  m_rotationQueue.begin();
       s != m_rotationQueue.end();
       s++)
  {
    // Renumber the reads in the slice to be read' as necessary
    SliceData * slice = *s;

    if (m_linear)
    {
      unsigned short depth = slice->getDepth();
      string newReadString;
      bool readChange = false;

      for (unsigned short i = 0; i < depth; i++)
      {
        const ReadData * rd = slice->getRead(i);
        string seqid = rd->getId();
        string newseqid = seqid + "'";

        if (i) { newReadString += " "; }

        if (m_readManager->find(newseqid))
        {
          // This read exists in the manager as read'
          newReadString += newseqid;
          readChange = true;
        }
        else
        {
          newReadString += seqid;
        }
      }

      if (readChange)
      {
        slice->resetReadId();
        slice->addReadId(newReadString.c_str());
      }
    }

    printSlice(slice);
    delete (slice);
  }
    
  m_out << "  </SliceRange>" << endl;
  m_out << "</Request>"      << endl;

  return m_errorCount;
}


void SliceRotator::handleOptions(Foundation * tf)
{
  tf->getOptions()->addOptionResult("n=i", &m_rotateSize,
                                    "Specify number of slices to rotate");

  tf->getOptions()->addOptionResult("L", &m_linear,
                                    "Specify linear rotation (renumber reads)");



  RequestManager::handleOptions(tf);
}
