#include "TilingManager.hh"
#include <stdio.h> //cuserid

TilingManager::TilingManager() : RequestManager("TilingManager")
{
  m_contigCount = 0;
  m_consensusSequence = NULL;
}

string TilingManager::m_asmblId("");
string TilingManager::m_caId("");
string TilingManager::m_user("");
string TilingManager::m_comment("");
string TilingManager::m_method("");
string TilingManager::m_mod_date("");
int TilingManager::m_doQuality(0);

TilingManager::~TilingManager()
{
  if (m_consensusSequence) { delete m_consensusSequence; }
}

void TilingManager::addSliceRange(long & index, long & gindex)
{
  if (m_consensusSequence)
  {
    m_consensusSequence->printTasm(m_asmblId, 
                                   m_user, 
                                   m_comment, 
                                   m_method, 
                                   m_caId, 
                                   m_mod_date,
                                   m_doQuality);
    delete m_consensusSequence;
    m_consensusSequence = NULL;

    m_out << "|" << endl;
  }

  m_consensusSequence = new ConsensusSequence(m_contigCount, index, m_out);
  
  m_contigCount++;

  index = gindex = 0L;
}

void TilingManager::addSlice(SliceData * slice)
{
  RequestManager::addSlice(slice);
  m_consensusSequence->addSlice(slice);
  delete slice;
}


int TilingManager::printResults()
{
  if (!m_errorCount && m_consensusSequence)
  {
    m_consensusSequence->printTasm(m_asmblId, 
                                   m_user, 
                                   m_comment, 
                                   m_method, 
                                   m_caId,
                                   m_mod_date,
                                   m_doQuality);
  }

  return m_errorCount;
}

void TilingManager::setAsmblId(const string & asmblId)
{
  m_asmblId = asmblId;
}

void TilingManager::setCAId(const string & caId)
{
  m_caId = caId;
}

void TilingManager::setUser(const string & user)
{
  m_user = user;
}

void TilingManager::setComment(const string & comment)
{
  m_comment = comment;
}

void TilingManager::setMethod(const string & method)
{
  m_method = method;
}


void TilingManager::handleOptions(Foundation * tf)
{
  RequestManager::addStandardOptions(tf);

  tf->getOptions()->addOptionResult("asmbl_id|i=s", &m_asmblId,
                                    "Specify the assembly id");

  tf->getOptions()->addOptionResult("ca_id|I=s", &m_caId,
                                    "Specify the ca_contig_id");

  tf->getOptions()->addOptionResult("method|d=s", &m_method,
                                    "Specify the method string");

  tf->getOptions()->addOptionResult("user|u=s", &m_user,
                                    "Specify the user (ed_pn string)");

  tf->getOptions()->addOptionResult("comment|n=s", &m_comment,
                                    "Specify the comment string");

  tf->getOptions()->addOptionResult("date|T=s", &m_mod_date,
                                    "Specify the mod_date");

  tf->getOptions()->addOptionResult("q", &m_doQuality,
                                    "Compute Quality Classes");

  tf->handleStandardOptions();

  loadConfig(tf->getConfig());
  RequestManager::handleOptionResult();

  // Set sensible defaults
  if (m_asmblId.empty()) { m_asmblId = "0";           }
  if (m_caId.empty())    { m_caId = m_asmblId;        }
  if (m_method.empty())  { m_method   = "slice2tasm"; }
  if (m_comment.empty()) { m_comment  = "CA FREE: 0"; }

  if (m_user.empty())     
  { 
    if (struct passwd * userInfo = getpwuid(geteuid()))
    {
      m_user = (string) userInfo->pw_name;
    }
    else
    {
      m_user = "Unknown";
    }
  }

  /*
  printConfig(asmbl_id, method, user, comment, output,
              doConsensus, doAmbiguity, highQualityThreshold);
  */
}

void TilingManager::loadConfig(ConfigFile * configFile)
{
  if (configFile)
  {
    ConfigSection * root = configFile->getRootSection();

    string var;
    while ((var = root->getNextVariable()) != "")
    {
      string val = root->getValue(var);
      
      // command line is authoritative for these

      if      (var == "asmbl_id" && m_asmblId.empty())  { m_asmblId = val;  }
      else if (var == "ca_id"    && m_caId.empty())     { m_caId = val;     }
      else if (var == "method"   && m_method.empty())   { m_method = val;   }
      else if (var == "ed_pn"    && m_user.empty())     { m_user = val;     }
      else if (var == "mod_date" && m_mod_date.empty()) { m_mod_date = val; }
      else if (var == "user"     && m_user.empty())     { m_user = val;     }
      else if (var == "comment"  && m_comment.empty())  { m_comment = val;  }
      else if (var == "output"   && m_outfile.empty())  { m_outfile = val;  }
      else
      {
        int ival = atoi(val.c_str());

        if      (var == "doConsensus") 
        { 
          m_doConsensus = m_doConsensus || ival; 
        }
        else if (var == "doAmbiguity") 
        { 
          m_doAmbiguity = m_doAmbiguity || ival; 
        }
        else if (var == "highQuality" && !m_highQualityThreshold)
        { 
          m_highQualityThreshold = ival; 
        }
      }
    }
  }
}

void TilingManager::printConfig()
{
  cout << "asmbl_id="    << m_asmblId              << endl;
  cout << "ca_id="       << m_caId                 << endl;
  cout << "method="      << m_method               << endl;
  cout << "user="        << m_user                 << endl;
  cout << "comment="     << m_comment              << endl;
  cout << "output="      << m_outfile              << endl;
  cout << "doConsensus=" << m_doConsensus          << endl;
  cout << "doAmbiguity=" << m_doAmbiguity          << endl;
  cout << "highQuality=" << m_highQualityThreshold << endl;
  cout << "mod_date="    << m_mod_date             << endl;
}

